(* But : outils de base pour traiter la logique MILL
   implmentation de : quelques transformations, conversion
   d'une formule de MILL, vrification avec un certain nombre
   d'abstractions *)
(* Note personnelle : une rcriture d'un programme permet
   la clarification de celui-ci *)
(* "Release Note" : pas de rgles libralises pour l'instant.
   Je pourrais ainsi mesurer l'impact
     a) du filtrage sur le nombre de variables
     b) des rgles libralises
     c) du filtrage sur l'interprtation des labels en CL
*)


(* Module : vrifier des formules de la logique classique.
   Mthode utilise : arbres de Beth (efficace et facile  programmer) *)


module SortedInt =
  struct type t = int let cmp a b = a - b end;;

module SortedDList =
  struct 
    type t = (int list * int list) 
    let cmp a b = if a < b then -1 else if a = b then 0 else 1
  end 

module type EltType = 
  sig
    type t
    val cmp: t -> t -> int
  end
   
module Helpers = 
  struct
    let is_in a = List.exists (fun x -> x = a)
    let rec union a = function
      []   -> a
    | t::q -> if is_in t a then union a q
                           else t::union a q

    let rec all_assoc elt = function
      []        -> []
    | (a, b)::c -> if a = elt then b::all_assoc elt c
                              else all_assoc elt c

    let rec make_single_list = function
       []  -> []
    | t::q -> if is_in t q then make_single_list q
                           else t::make_single_list q
                  
    module SortedList = functor (Elt: EltType) ->
      struct
        type t = Elt.t
        let cmp = Elt.cmp

        let rec insert (e: t) = function
          ([]: t list) -> [e]
        | (a::b)       -> if cmp a e = 0 then a::b
                            else if cmp a e < 0 then a::(insert e b)
                            else e::a::b
        
        let merge a b =
          let rec merge_uc = function
            ([], e)          -> e
          | (e, [])          -> e
          | ((a::b), (c::d)) -> if cmp a c < 0
                                  then a::merge_uc (b, (c::d))
                                  else if cmp a c = 0
                                    then a::merge_uc (b, d)
                                    else c::merge_uc ((a::b), d)
        in merge_uc (a, b)

        (* Removes b from a *)
        let rec remove (a: t list) (b: t list) = match a with
          []    -> []
        | x::y  -> if is_in x b then remove y b else x::remove y b

        let concat (a: t list) (b: t list) =
          let rec concat_uc = function
            ([], tt)         -> tt
          | (tt, [])         -> tt
          | ((a::b), (c::d)) -> if cmp a c < 0
                                  then a::concat_uc (b, (c::d))
                                  else c::concat_uc ((a::b), d)
          in concat_uc (a, b)

        (* peut-tre source d'erreurs *)
        let rec sub_part = function
          []     -> [[]]
        | a::b   -> let t = sub_part b in
                      make_single_list (union (List.map (fun x -> a::x) t) t)
          
    end    
  end;;
  
module Proof_cl = 
  struct
    type variable = int
    type formule = V of variable | Impl of formule * formule
                 | Not of formule | Or of formule * formule
                 | And of formule * formule
    exception Contradiction
    let has_a_proof phi =
      let rec make_tableau lp ln = function
        (V i, true)  -> if Helpers.is_in i ln 
                          then raise Contradiction
                          else (i::lp, ln)
      | (V i, false) -> if Helpers.is_in i lp
                          then raise Contradiction
                          else (lp, i::ln)
      | (Impl (a, b), true) ->
                        (try make_tableau lp ln (a, false)
                          with Contradiction -> make_tableau lp ln (b, true))
      | (Impl (a, b), false) ->
                        let (llp, lln) = make_tableau lp ln (a, true)
                          in make_tableau llp lln (b, false)
      | (Or (a, b), true) ->
                        (try make_tableau lp ln (a, true)
                          with Contradiction -> make_tableau lp ln (b, true))
      | (Or (a, b), false) ->
                        let (llp, lln) = make_tableau lp ln (a, false)
                          in make_tableau llp lln (b, false)
      | (And (a, b), true) ->
                        let (llp, lln) = make_tableau lp ln (a, true)
                          in make_tableau llp lln (b, true)
      | (And (a, b), false) ->
                        (try make_tableau lp ln (a, false)
                          with Contradiction -> make_tableau lp ln (b, false))
      | (Not a, s) -> make_tableau lp ln (a, not s)
    in try let _ =  make_tableau [] [] (phi, false) in false
       with Contradiction -> true
  end;;

(* Module : traitement des graphes *)

module IntSList = Helpers.SortedList(SortedInt);;

module Graph = 
struct
  module Make =      
    functor (Elt : EltType) ->
       struct
         type elt = Elt.t
         type assoc_list = elt * elt list
         type demi_graph = assoc_list list
         type t = demi_graph * demi_graph

         let cmp = Elt.cmp

         module ListElt = Helpers.SortedList(Elt)
       
         let im_followers (e: elt) (a: demi_graph) =
           try List.assoc e a with
             Not_found -> []
             
         let followers (x: elt) (g: demi_graph) =
           let step = ref [x] and ret = ref [x] in
             begin
               while !step <> [] do
                 step := List.flatten
                   (List.map (fun x -> im_followers x g) !step);
                 ret := (!step) @ (!ret);
               done;
               !ret
             end
           
         let after     (x: elt) (g: t) = followers x (snd g)
         let before    (x: elt) (g: t) = followers x (fst g)
         let im_after  (x: elt) (g: t) = im_followers x (snd g)
         let im_before (x: elt) (g: t) = im_followers x (fst g)

         let rec half_add_item (x: elt) = function
           ([]: demi_graph) -> [x, []]
         | (a::b)           -> if cmp (fst a) x < 0
                                 then a::(half_add_item x b)
                                 else if cmp (fst a) x = 0
                                   then a::b
                                   else (x, [])::a::b

         let add_item (x:elt) (g:t) = 
           ((half_add_item x (fst g), half_add_item x (snd g)) : t)

         let half_add_link ((e1, e2): elt * elt) (g: demi_graph) =
           if e1 = e2 then (half_add_item e1 (half_add_item e2 g))
           else
             let rec func1 = function
                []        -> failwith "Graph.half_add_link"
              | (a, b)::c -> if cmp a e1 < 0
                               then (a, b)::func1 c
                               else (a, (ListElt.insert e2 b))::c
             in
               func1 (half_add_item e2 (half_add_item e1 g))
         
         let add_link ((x, y):elt*elt) (g: t) =
           ((half_add_link (x, y) (fst g), half_add_link (y, x) (snd g)): t)

         
       end
end;;

module EltLabel = struct
  type t = int list
  let cmp a b = if a < b then -1 else if a > b then 1 else 0
end;;

module LabelGraph = Graph.Make(EltLabel);;

module LabelWork =
struct 
  type t = LabelGraph.t
  type elt = EltLabel.t

  let cmp = EltLabel.cmp
  
  let partitions (e: elt) =
    let rec parties = function
      ([]: elt)   -> [([]:elt)]
    | a::b        -> let p = parties b in
                       (List.map (fun x -> (a::x)) p) @ p
    in let p = parties e in
      List.map2 (fun x -> fun y -> (x, y)) p (List.rev p)

  let find_decomps_1 (g: t) (e: elt) =
    let pred (x, y) = not ((x = []) or (y = [])) in
    let suite = LabelGraph.after e g in
      ((List.sort cmp (List.flatten (List.map partitions suite)))
         : (elt * elt) list)

  let rec de_cut = function
    ((a::b), (c::d)) -> if c < a
                          then raise Not_found
                          else if a = c then de_cut (b, d)
                                        else a::(de_cut (b, (c::d)))
  | (t, [])          -> t
  | ([], t)          -> raise Not_found
      
  let find_decomps_2 (g: t) (e: elt) = 
    let ll = List.map fst (fst g) in
      List.flatten (List.map (
        fun x -> try let d = de_cut (x, e) in
          if Helpers.is_in x ll
            then [d, x]
            else []
                 with Not_found -> [] )
        ll);;

  let solve_yx y x = 
    let rec interm y x prof =
      try let a = de_cut (y, x) in
        (a, prof)::interm a x (prof + 1)
      with Not_found -> []
    in if x = [] then [y, 1] else interm y x 1

  let rec substitute (b: elt) ((a: int list), i) =
    if i = 0 then a
             else IntSList.concat b (substitute b (a, (i-1)))
end;;




(* Programme principal *)


let debug_mode = ref false;;
let count_call = ref 0;;

module IntCListSList = Helpers.SortedList(SortedDList);;

let nb_id = ref 01;;
let new_id () = let a = !nb_id
  in begin incr nb_id; a end;;

module Mill =
  struct
    type variable = int
    type formule =
      V of variable | Impl of formule * formule | And of formule * formule
    
(* Deux interprtations diffrentes d'une formule :
   la premire, le nombre d'utilisations de chaque variable
   la deuxime, le plongement dans la logique classique *)
      
    let rec to_cl = function
      V i         -> Proof_cl.V i
    | Impl (a, b) -> Proof_cl.Impl (to_cl a, to_cl b)
    | And (a, b)  -> Proof_cl.And (to_cl a, to_cl b)

    let has_a_proof_in_cl a =
      Proof_cl.has_a_proof (to_cl a)

    let rec valuation a e =
      match a with
        V i         -> if i = e then 1 else 0
      | Impl (a, b) -> (valuation b e) - (valuation a e)
      | And (a, b)  -> (valuation a e) + (valuation b e)

    let rec occur_variables = function
      V i         -> [i]
    | Impl (a, b) -> IntSList.merge (occur_variables a)
                                    (occur_variables b)
    | And (a, b)  -> IntSList.merge (occur_variables a)
                                    (occur_variables b)

    let valuation phi = 
      List.map (fun x -> (x, valuation phi x)) (occur_variables phi)

    let string_of_proposition p = 
      let rec sof a = match a with 
          And (p1, p2) -> "(" ^ sof (p1) ^ "*" ^ sof (p2) ^ ")"
        | Impl (p1, p2) -> "(" ^ sof (p1) ^ "-*" ^ sof (p2) ^ ")"
	| V i -> Char.escaped (Char.chr i)
      in sof p

  end;;
  
module AnnotationLitt = 
  struct
    (* On associe  une constante la liste des poids de chaque variable *)
    type annotations = (int * ((int * int) list)) list

    let rec clean_valuation = function
      []                -> []
    | ((a, (b:int))::c) -> if b = 0 then clean_valuation c
                                    else (a, b)::clean_valuation c
    
    let same_valuation (a: (int * int) list) (b: (int * int) list) (sg: int) =
      let rec fun1 = function
        []          -> true
      | (t1, t2)::q -> try
                         if (List.assoc t1 a) = (sg * t2) then
                           fun1 q
                         else false
                       with Not_found -> false
      in
        let (pa, pb) = (clean_valuation a, clean_valuation b) in
          if (List.length pa) <> (List.length pb) then false
          else fun1 pb

    (* Simplement on a besoin d'ajouter des annotations et de
       filtrer une liste de 'a * Mill.formule * int list
       pour placer celles qui vrifient [$2]=[$3] en tte
       (et, si l'utilisateur le veut bien, on les supprime) *)

    let bool2int = function true -> 1 | false -> -1
    
    let rec add_valuations (a: (int * int) list) (b: (int * int) list) =
      let app = Helpers.union (List.map fst a) (List.map fst b) in
        let rec f = function
          []   -> []
        | t::q -> (t, ((try List.assoc t a with Not_found -> 0)
                     + (try List.assoc t b with Not_found -> 0)))::f q
      in f app
    
    let rec valuation_of_label (a: annotations) = function
      []   -> []
    | t::q -> add_valuations (List.assoc t a) (valuation_of_label a q)
   
    (* Note sur les arguments de la fonction can_instantiate :
       a -> pas de problmes
       b -> la formule en cours de traitement
       c -> le label en question
       sg -> true ou false, le signe de la formule en cours de traitement *)
    let can_instantiate 
       (a: annotations) 
       (b: Mill.formule) (c: int list) (sg: bool) =
          same_valuation (Mill.valuation b) (valuation_of_label a c) 
            (bool2int (not sg))
        
  end;;

(* Le gros du problme : les fonctions de cration / traitement
   des tableaux avec labels pour MILL *)

type class_op = Pi_alpha | Pi_beta | Ground;;

(* Reconnaissons un mrite  la version MILL : on n'a plus besoin
   des indentificateurs. En effet il n'y a plus de risque de branche
   redondante.
*)

type formula = Mill.formule;;

type label = int list;;

type signed_formula = bool * class_op * formula;;

type formula_tree =
  Leaf of int * bool
| Node of formula_tree * signed_formula * formula_tree;;

let formula_tree_of_formula = function (p: formula) ->
  let rec interne (v, x) = match (v, x) with
    (pol, Mill.V i) -> Leaf (i, pol)
  | (false, Mill.And (a, b))
                    -> Node (interne (false, a), 
                              (false, Pi_beta, x),
                             interne (false, b))
  | (false, Mill.Impl (a, b))
                    -> Node (interne (true, a), 
                              (false, Pi_alpha, x),
                             interne (false, b))
  | (true, Mill.And (a, b))
                    -> Node (interne (true, a), 
                              (true, Pi_alpha, x),
                             interne (true, b))
  | (true, Mill.Impl (a, b))
                    -> Node (interne (false, a), 
                              (true, Pi_beta, x),
                             interne (true, b))
  in interne (false, p);;

let exemple =
  Mill.Impl (
    Mill.And (
      Mill.V 1, 
      Mill.Impl (Mill.V 1, Mill.V 2)),
    Mill.V 2);;


module SortedIntList =
  struct
    type t = int list
    let cmp (a: int list) b = if a < b then -1 else if a > b then 1 else 0
  end;;

module IntListSList =
  Helpers.SortedList(SortedIntList);;

(* Module destin  la recherche des instantiations possibles 
   des diverses pi beta -- ici il n'y a que deux cas  traiter,
   contre trois dans le cadre de BI propositionnel tout entier *)
module Instantiations =
  struct
    (* La permire fonction traite du cas T* *)
    let rule_TMultM (graph: LabelGraph.t) (c1: int) (c2: int) (lab: int list) =
      let a = ref (Helpers.make_single_list 
                    (List.flatten 
                      (List.map (fun x -> LabelWork.solve_yx x lab)
                        (List.map fst (fst graph)))))
      and b = ref graph
        in
          begin
            while !a <> [] do
              let hda  = List.hd !a and
                  cij  = IntSList.merge [c1] [c2] in
              let yci  = LabelWork.substitute [c1] hda and
                  ycj  = LabelWork.substitute [c2] hda and
                  ycij = LabelWork.substitute cij hda and
                  yx   = LabelWork.substitute lab hda
              in
                a := List.tl !a;
                b := LabelGraph.add_item yci (LabelGraph.add_item ycj !b);
                b := LabelGraph.add_link (ycij, yx) !b;
                let c = ref (List.flatten
                              (List.map (fun x -> LabelWork.solve_yx x lab)
                                        (LabelGraph.im_after yx graph))) in
                  while !c <> [] do
                    let hdc  = List.hd !c in
                    let zci  = LabelWork.substitute [c1] hdc and
                        zcj  = LabelWork.substitute [c2] hdc and
                        zcij = LabelWork.substitute cij hdc
                    in
                      c := List.tl !c;
                      b := LabelGraph.add_link (ycij, zcij) !b;
                      b := LabelGraph.add_link (yci, zci) !b;
                      b := LabelGraph.add_link (ycj, zcj) !b;
              done;
          done;
          !b
        end

        
    let rule_FImplM (graph: LabelGraph.t) (c1: int) (lab: int list) =
      let aa = IntListSList.merge (LabelGraph.after lab graph)
                                  (LabelGraph.before lab graph) in
      let a = ref (Helpers.make_single_list (List.flatten
                                              (List.map IntSList.sub_part aa)))
      in let b = ref graph in
        begin
          while !a <> [] do
            let y = List.hd !a in
              let y = List.hd !a in
                let evz = ref (LabelGraph.im_after y graph) in
                  begin
                    b := LabelGraph.add_item (IntSList.concat y [c1]) !b;
                    while !evz <> [] do
                      b := LabelGraph.add_link
                        (IntSList.concat [c1] y,
                         IntSList.concat [c1] (List.hd !evz)) !b;
                      evz := List.tl !evz;
                    done;
                  end;
            a := List.tl !a;
          done;
          !b
        end
    
  end;;
    

(* Sous-fonctions de traitement des connections *)    
module Connections =
  struct
    exception Found

    let find_connections (pos: (formula * int list) list)
                         (neg: (formula * int list) list)
                         (graph: LabelGraph.t) =
      let rec paires_possibles pos = function
        []        -> false
      | (a, b)::q -> try
                       begin
                         let ss = ref (Helpers.all_assoc a pos) in
                           while !ss <> [] do
                             let s = List.hd !ss in
                               if Helpers.is_in s (LabelGraph.before b graph)
                                 then raise Found;
                             ss := List.tl !ss;
                           done;
                           paires_possibles pos q
                       end
                     with Found -> true
      in paires_possibles pos neg
  end;;

module Choice =
  struct
    let category (gr: LabelGraph.t) (cat: class_op)
                 (a: formula_tree) (l: int list) 
                 (pi_bt: (formula * int list) list) =
      match a with
        Node (_, (p, c, f), _) when c = cat
          -> if c = Pi_beta 
               then
                 if Helpers.is_in (f, l) pi_bt 
                   then []
                   else
                     [a, l, ([], [])]
             else [a, l, ([], [])]
      | Node (_, (p, c, f), _) -> []
      | Leaf (_) -> if cat = Ground then [a, l, ([], [])] else []

    let is_a_good_choice (a: AnnotationLitt.annotations)
                         (f: formula) (lg: int list) (ld: int list) 
                         =
      match f with
    | Mill.Impl (g, d) -> 
        (AnnotationLitt.can_instantiate a g lg false) &&
        (AnnotationLitt.can_instantiate a d ld false)
    | Mill.And (g, d) ->
        (AnnotationLitt.can_instantiate a g lg false) &&
        (AnnotationLitt.can_instantiate a d ld false)
    | _               -> failwith "This is no pi-beta rule"
     
    let get_formula = function
      Node (_, (_, _, f), _) -> f
    | Leaf (i, _)            -> Mill.V i

    exception Open_branch
   
    exception Made_a_choice of
      (formula_tree * (int list * int list)) *
      (formula_tree * int list * ((int list * int list) list * (int list * int
      list) list)) list * (int list);;
    
    (* Renvoie un couple choix fait / reste des pi_beta *)
    (* <APOLOGIES>Je suis dsol mes types sont illisibles </APOLOGIES>*)
    let choose_an_instantiation (a: AnnotationLitt.annotations)
                                (f: ((formula_tree * 
                                (int list) * 
                                ((int list * int list) list 
                                  * (int list * int list) list))) list) =
      try 
        if f = [] then failwith "no choice: the list is empty."
        else
          let cf = ref f and bf = ref [] in
            begin
              while !cf <> [] do
                begin
                  let (phi, lab, (to_do, tdone)) = List.hd !cf in
                    let cto_do = ref to_do and bto_do = ref [] in
                      while !cto_do <> [] do
                        if is_a_good_choice a (get_formula phi) 
                           (fst (List.hd !cto_do)) (snd (List.hd !cto_do))
                        then
                          raise (Made_a_choice ((phi, List.hd !cto_do),
                            (!bf)@(List.tl !cf)@[phi, lab, 
                              ((List.tl !cto_do)@(!bto_do), 
                               ((List.hd !cto_do)::tdone))], lab))
                        else
                          bto_do := (List.hd !cto_do)::(!bto_do);
                          cto_do := List.tl !cto_do;
                      done;
                  bf := (List.hd !cf)::(!bf);
                  cf := List.tl !cf;
                end
              done;
              raise Open_branch;
            end
      with Made_a_choice (a, b, c) -> (a, b, c)

    (* Cette fonction remplit les listes des instantiations possibles.
       Elle se sert donc du graphe des dpendances. *)
    let find_all_possible_instantiations 
       (g: LabelGraph.t)
       (a: AnnotationLitt.annotations)
       (f: ((formula_tree *
             (int list) *
             ((int list * int list) list *
              (int list * int list) list))) list) =
       let cf = ref f and bf = ref [] in
         begin
           while !cf <> [] do
             let (phi, lab, (_, tdone)) = List.hd !cf in
               match phi with
                 Node (_, (b, _, _), _)
                   -> if b 
                        then 
                          begin (* T-* *)
                            bf := (phi, lab, 
                                    (IntCListSList.remove
                                      (LabelWork.find_decomps_2
                                                   g
                                                   lab) tdone,
                                       tdone))::!bf;
                            cf := List.tl !cf;
                          end
                        else
                          begin (* F* *)
                            bf := (phi, lab,
                                     (IntCListSList.remove
                                       (LabelWork.find_decomps_1
                                                    g
                                                    lab) tdone,
                                        tdone))::!bf;
                            cf := List.tl !cf;
                          end
               | _ -> failwith "pi-beta is not a node."
           done;
           !bf
         end

  end;;


(* Le module prsidant  la construction d'un tableau.
   Il contient la fonction principale. *)
module Tableau =
  struct
    exception Open_branch
    
    let rec step_in_tableau (pos: (formula * int list) list)
                            (neg: (formula * int list) list)
                            (graphe: LabelGraph.t)
                            (gr: (formula_tree * (int list) * ('a * 'b)) list)
                            (pi_a: (formula_tree * (int list) * ('a * 'b)) list)
                            (pi_b: (formula_tree * (int list) * ('a * 'b)) list)
                            (pibt: (formula * label) list) 
                            (a: AnnotationLitt.annotations) =
    begin if !debug_mode then incr count_call; 
      if Connections.find_connections pos neg graphe 
        then () (* Cas favorable *)
        else
          if gr <> [] then
            match List.hd gr with
              (Leaf (var, pol), lab, _) ->
                if pol then 
                  step_in_tableau ((Mill.V var, lab)::pos) neg graphe
                                  (List.tl gr) pi_a pi_b pibt a
                       else
                  step_in_tableau pos ((Mill.V var, lab)::neg) graphe
                                  (List.tl gr) pi_a pi_b pibt a
            | _ -> failwith "ground item is not a leaf"
          else if pi_a <> [] then
            (* Il faut recalculer le graphe des dpendances ici *)
            let (ngr, lg, ld, fg, fd, new_a) = 
              match List.hd pi_a with
                (Node (fg, (pol, _, Mill.And (ffg, ffd)), fd), lab, (_, _)) ->
                   let (c2, c1) = (new_id(), new_id()) in
                     (Instantiations.rule_TMultM graphe c1 c2 lab,
                       [c1], [c2], fg, fd,
                       (c1, Mill.valuation ffg)::(c2, Mill.valuation ffd)::a
                     )
              | (Node (fg, (pol, _, Mill.Impl (ffg, ffd)), fd), lab, (_, _)) ->
                   let c1 = new_id() in
                     (Instantiations.rule_FImplM graphe c1 lab,
                       [c1],
                       IntSList.insert c1 lab, fg, fd,
                       (c1, Mill.valuation ffg)::a
                     )
              | _ -> failwith "pi-alpha is not a node"
            in
              step_in_tableau pos neg ngr
                ((Choice.category graphe Ground fg lg pibt)@
                 (Choice.category graphe Ground fd ld pibt))
                ((Choice.category graphe Pi_alpha fg lg pibt)@
                 (Choice.category graphe Pi_alpha fd ld pibt)@(List.tl pi_a))
                ((Choice.category graphe Pi_beta fg lg pibt)@
                 (Choice.category graphe Pi_beta fd ld pibt)@pi_b) pibt new_a
           else if pi_b = [] then raise Open_branch
           else (* Il reste des pi_beta  traiter. *)
                (* Essayons de faire plus propre que dans la premire version
                   qui n'tait pas ma foi un modle du genre... *)
             let npi_b = Choice.find_all_possible_instantiations
                                  graphe a pi_b
               in
                 let (choix, rpi_b, label) = 
                   Choice.choose_an_instantiation a npi_b
                 in
                   match choix with
                     (Node (fg, (_, _, f), fd), (lg, ld))
                       -> begin 
                          step_in_tableau pos neg graphe
                            (Choice.category graphe Ground fg lg pibt)
                            (Choice.category graphe Pi_alpha fg lg pibt)
                            ((Choice.category graphe Pi_beta fg lg pibt)@rpi_b)
                           ((f, label)::pibt) a;
                          step_in_tableau pos neg graphe
                            (Choice.category graphe Ground fd ld pibt)
                            (Choice.category graphe Pi_alpha fd ld pibt)
                            ((Choice.category graphe Pi_beta fd ld pibt)@rpi_b)
                           ((f, label)::pibt) a;
                           end
                   | _ -> failwith "unexpected: received Leaf / awaited Node"
    end

    let find_a_proof f =
    try
      if not (*Mill.has_a_proof_in_cl f*) true
        then failwith "this is false in CL hence cannot be true in MILL"
      else
      if not (AnnotationLitt.same_valuation [] (Mill.valuation f) 1)
      then failwith "this formula is unbalanced"
      else
        let phi = formula_tree_of_formula f in
          match phi with
            Leaf _ -> failwith "an atom is never provable"
          | Node (_, (_, cat, _), _) ->
                  match cat with
                    Pi_alpha -> begin
                                  try
                                    let _ = step_in_tableau
                                              [] [] ([],[]) [] 
                                              [phi, [], ([],[])]
                                              [] [] []
                                    in true
                                  with Choice.Open_branch -> false
                                  |    Open_branch        -> false
                                end
                  | Pi_beta -> begin
                                  try
                                    let _ = step_in_tableau
                                              [] [] ([],[]) [] [] 
                                              [phi, [], ([],[])] [] []
                                    in true
                                  with Choice.Open_branch -> false
                                  |    Open_branch        -> false
                                end
                  | _         -> failwith "unexpected error"
         with
          Failure s -> let _ = 
                              print_string 
                                ("failure: " ^ s ^ "\n")
                             in false
        
                                  
  end;;


let rec p_left i j = 
  if i = 0 then Mill.V j
  else Mill.And (Mill.V j, p_left (i-1) (j+1));;
  

let rec p_right i j = 
  if i = 0 then Mill.V j
  else Mill.And (p_right (i-1) (j+1), Mill.V j);;

type injonction = No_operation
                | Check of formula
                | Stat
                | Help
                | Strategy of bool
                | Syntax_error ;;
